/* $Id: neb-hdtoolbox.c,v 1.32 2005/05/16 14:08:25 stoehr Exp $ */

/*
 * neb-hdtoolbox - An Amiga harddisk toolbox utility for NetBSD/amiga
 *
 * Written by Florian Stoehr <netbsd@wolfnode.de>
 */

#include <sys/disklabel.h>
#include <sys/ioctl.h>
#include <sys/param.h>

#include <arpa/inet.h>

#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <util.h>

#include <signal.h>

struct amigapart; /* Forward */

void usage(void);
char * ReadStr(void);
uint32_t ReadUL(void);
void ResetRDB(void);
void LoadLabel(void);
void GetDefaultLabel(void);
void PrintDiskInfo(void);
void CmdLoop(void);
struct amigapart * AddPartition(char, struct amigapart *);
int EditPartition(struct amigapart *);
void DeletePartition(struct amigapart *);
void ArrangeBlocks(void);
int32_t AmiChecksum(unsigned char *);
void CalcChecksums(void);
void FormatToAmiga(void);
void FormatRDBToLocal(void);
void FormatPARTToLocal(struct amigapart *ap);
void FormatToLocal(void);
void Write(uint32_t, uint32_t, uint32_t blockstep);
int IsAreaUsed(uint32_t, uint32_t, struct amigapart *);
void M_PrintHelp(void);
void M_ClearRDB(void);
void M_PrintPart(void);
void M_DeletePart(void);
void M_Write(void);
void MS_EditPartition(char);
void P_FmtLine(int, uint32_t, uint32_t, uint32_t, const char *, \
	const char *);
void P_WholeDisk(void);
int main(int argc, char *argv[]);

#define SHORT(p) ((p)[0]<<8 | (p)[1])
#define LONG(p) (SHORT(p)<<16 | SHORT(p+2))
#define BETWEEN(p,a,b) ((p >= a) && (p <= b))
#define SPAN(p,q,a,b) ((p <= a) && (q >= b))

static const char *rd_id = "RDSK";
static const char *pa_id = "PART";
static const char *pa_dos = "NBR";
static const uint32_t ND = 0xffffffff;

static char buffer[MAXPATHLEN];
char *s_Device;
char *s_FullDev;
struct disklabel *lp;
struct rdb *rdbp;
int i_FD;

/*
 * Menu commands
 */
static const struct interactive
{
	const char *cmd;
	const char *hlptext;
	void (*ptr)(void);
} interactive[] =
{
	{"?", "This screen", M_PrintHelp},
	{"C", "Clear (erase) RDB", M_ClearRDB},
	{"P", "Print partitions", M_PrintPart},
	{"D", "Delete partition", M_DeletePart},
	{"W", "Write data to disk", M_Write},
	{"Q", "Quit program", NULL},
	{"[a-e]", "Edit named partition", NULL},
	{NULL, NULL, NULL}
};

/* 
 * Structure containing the partition data.
 * There exists one "PART" block per partition.
 */
typedef struct amigapart_data
{
	char id[4];
	uint32_t size_in_longs;
	uint32_t checksum;
	uint32_t hostid;
	uint32_t next;
	uint32_t flags;
	uint32_t reserved_1[2];
	uint32_t devflags;
	char drivenamelen;
	char drivename[31];
	uint32_t reserved_2[15];
	
	// DOS Environment Vector
	uint32_t size_of_vector;
	uint32_t sizeblock;
	uint32_t secorg;
	uint32_t surfaces;
	uint32_t sectors_block;
	uint32_t blocks_track;
	uint32_t reserved;
	uint32_t prealloc;
	uint32_t interleave;
	uint32_t lowcyl;
	uint32_t highcyl;
	uint32_t numbuffer;
	uint32_t bufmemtype;
	uint32_t maxtransfer;
	uint32_t mask;
	uint32_t bootpri;
	char dostype[4];
	uint32_t baud;
	uint32_t control;
	uint32_t bootblocks;
	uint32_t reserved_4[12];
} amigapart_data;

/*
 * Internal representation of the amigapart
 */
typedef struct amigapart
{
	amigapart_data data;
	struct amigapart *next;
	struct amigapart *prev;
} amigapart;

/*
 * Structure containing the RDB data as written on disk
 */
typedef struct rdb_data
{
	// RDB base information
	char id[4];
	uint32_t size_in_longs;
	int32_t checksum;
	uint32_t hostid;
	uint32_t block_size;
	uint32_t flags;
	uint32_t bad_blocklist;
	uint32_t partitionlist;
	uint32_t filesyshdrlist;
	uint32_t driveinit_code;
	uint32_t reserved_1[6];
	
	// Physical drive characteristics
	uint32_t cylinders;
	uint32_t sectors;
	uint32_t heads;
	uint32_t interleave;
	uint32_t parking_zone;
	uint32_t reserved_2[3];
	uint32_t writeprecomp;
	uint32_t reducedwrite;
	uint32_t steprate;
	uint32_t reserved_3[5];

	// Logical drive characteristics
	uint32_t rdb_blocklo;
	uint32_t rdb_blockhi;
	uint32_t locylinder;
	uint32_t hicylinder;
	uint32_t cylblocks;
	uint32_t autoparkseconds;
	uint32_t highrsdkblock;
	uint32_t reserved_4;
	
	// Drive identification
	char diskvendor[8];
	char diskproduct[16];
	char diskrevision[4];
	char controllervendor[8];
	char controllerproduct[16];
	char controllerrevision[4];
	uint32_t reserved_5[10];
} rdb_data;

/*
 * Internal representation of the RDB
 */
typedef struct rdb
{
	rdb_data data;
	amigapart *partitionlist_lnk;
} rdb;


/*
 * Print usage information, then exit program
 */
void
usage(void)
{
	printf("Usage is: neb-hdtoolbox device\n"\
	       "          where device is the desired disk, e.g. \"sd0\"\n"); 
			   
	exit(EXIT_FAILURE);
}
 
/*
 * Read a string from stdin
 */
char *
ReadStr(void)
{
	fgets(buffer, MAXPATHLEN, stdin);
	
	return strdup(buffer);
}

/*
 * Read a 32 bit (unsigned) integer from stdin
 */
uint32_t
ReadUL(void)
{
	uint32_t u;	
	char *s;
	
	s = ReadStr();	
	sscanf(s, "%d", &u);
	free(s);
	
	return u;
}

/*
 * Reset a RDB. This will also delete all partitions.
 */
void 
ResetRDB(void)
{
	/* Remove all PART blocks in the chain */
	amigapart *cp, *cp_n;
	uint32_t numrescyl;
	
	/* Cleanup old data */
	cp = rdbp->partitionlist_lnk;
	cp_n = NULL;
	
	while (cp != NULL) {
		cp_n = cp->next;
		free(cp);
		cp = cp_n;
	}
	
	/* Reset partlist entry point */
	rdbp->partitionlist_lnk = NULL;
	
	/* Cleanup data block */
	memset(&rdbp->data, 0x00, sizeof(rdb_data));
	
	/* 
	 * Calculate number of reserved cylinders. We need at least
	 * 500 KB, so calc the number of needed cylinders based on
	 * blocks/cylinder and bytes/block.
	 * 
	 * Btw, I know 512 = 0x200 = 0 in the following case ;-)
	 */
	numrescyl = ceil((float)(512 * 1024) / (float)(0x200 * lp->d_secpercyl));
	
	/* Fill in default values */
	memcpy(rdbp->data.id, rd_id, 4);
	rdbp->data.size_in_longs = 0x40;
	rdbp->data.hostid = 7;
	rdbp->data.block_size = 0x200;
	rdbp->data.flags = 0x17;
	rdbp->data.bad_blocklist = ND;
	rdbp->data.partitionlist = ND;
	rdbp->data.filesyshdrlist = ND;
	rdbp->data.driveinit_code = ND;
	rdbp->data.reserved_1[0] = ND;
	rdbp->data.reserved_1[1] = ND;
	rdbp->data.reserved_1[2] = ND;
	rdbp->data.reserved_1[3] = ND;
	rdbp->data.reserved_1[4] = ND;
	rdbp->data.reserved_1[5] = ND;
	rdbp->data.cylinders = lp->d_ncylinders;
	rdbp->data.sectors = lp->d_nsectors;
	rdbp->data.heads = lp->d_ntracks;
	rdbp->data.interleave = lp->d_interleave;
	rdbp->data.rdb_blocklo = 0;
	rdbp->data.rdb_blockhi = (lp->d_secpercyl * numrescyl) - 1; /* two cylinders */
	rdbp->data.locylinder = numrescyl;
	rdbp->data.hicylinder = lp->d_ncylinders - 1; /* whole disk */
	rdbp->data.cylblocks = lp->d_secpercyl;
	strcpy(rdbp->data.diskvendor, "NetBSD");
	strcpy(rdbp->data.diskproduct, lp->d_typename);
}

/*
 * Load the current RDSK and PART chain. We're using our own mechanism
 * for loading an existing label here since we (in a later version) want
 * to rewrite all label parts, not only the NetBSD ones here. The "load
 * disklabel" call is inappropriate here.
 */
void 
LoadLabel(void)
{
	amigapart *newap, *curap;
	uint32_t nextblock, numrescyl;
	int cur_offs, i;
	unsigned char *blbuf;
	
	i_FD = opendisk(s_Device, O_RDONLY, buffer, MAXPATHLEN, 0);
	
	if (i_FD == -1)	{	
		printf("ERROR opening device. Make sure you are root.\n");

		exit(EXIT_FAILURE);
	}
	
	/*
	 * Load the first 16 blocks. RDSK must exist within those.
	 */
	blbuf = (unsigned char *)malloc(512 * 16);

	if (read(i_FD, blbuf, 512*16) == -1)	{
		perror("ERROR reading from disk");

		exit(EXIT_FAILURE);
	}
	
	/*
	 * Scan for a valid RDSK within each block
	 */
	for (i = 0; i < 16; i++) {
		if (
			(*(blbuf + (i * 512)) == 'R') &&
			(*(blbuf + (i * 512 + 1)) == 'D') &&
			(*(blbuf + (i * 512 + 2)) == 'S') &&
			(*(blbuf + (i * 512 + 3)) == 'K')) {
			
			/* RDSK found */
			memcpy(&rdbp->data, (blbuf + (i * 512)), 512);
			
			/* Transform to local endianness */
			FormatRDBToLocal();			
			
			/* 
			 * Overwrite the lp-data of the default disklabel;
			 * this *MIGHT* differ if the rdsk wasn't created
			 * with neb-hdtoolbox.
			 */
			lp->d_ncylinders = rdbp->data.cylinders;
			lp->d_nsectors = rdbp->data.sectors;
			lp->d_ntracks = rdbp->data.heads;
			lp->d_interleave = rdbp->data.interleave;
			lp->d_secpercyl = rdbp->data.cylblocks;

			/* 
			 * Calculate number of reserved cylinders. We need at least
			 * 500 KB, so calc the number of needed cylinders based on
			 * blocks/cylinder and bytes/block.
			 * 
			 * Btw, I know 512 = 0x200 = 0 in the following case ;-)
			 */
			numrescyl = ceil((float)(512 * 1024) / (float)(0x200 * lp->d_secpercyl));

			/*
			 * New RDSK will occupy at least 500 KB for the RDSK/PART
			 * and will use the whole disk.
			 */
			rdbp->data.rdb_blocklo = 0;
			rdbp->data.rdb_blockhi = (lp->d_secpercyl * numrescyl) - 1; /* two cylinders */
			rdbp->data.locylinder = numrescyl;
			rdbp->data.hicylinder = lp->d_ncylinders - 1; /* whole disk */
			
			/*
			 * Scan for existing PART sections
			 */			 
			nextblock = rdbp->data.partitionlist;
			curap = NULL;			
			
			while (nextblock != ND) {
				newap = (amigapart *)malloc(sizeof(amigapart));
				memset(newap, 0x00, sizeof(amigapart));
				
				if (lseek(i_FD, nextblock * 512, SEEK_SET) == -1) {
					perror("ERROR seeking partition offset");
					
					exit(EXIT_FAILURE);
				}
				
				/*
				 * We're operating on the device itself. Therefore,
				 * attempting not to read n*512 bytes (e.g. reading
				 * sizeof(amigapart_data)) will FAIL.
				 */				
				if (read(i_FD, blbuf, 512) == -1) {
					perror("ERROR reading partition data");
					
					exit(EXIT_FAILURE);
				}
				
				memcpy(&newap->data, blbuf, sizeof(amigapart_data));
				
				/*
				 * The data is always in amiga-endian format, so 
				 * convert to local endianness now
				 */
				FormatPARTToLocal(newap);
				
				/*
				 * See if the read partition is valid 
				 */			
				if (curap == NULL)
					rdbp->partitionlist_lnk = newap;
				else
					curap->next = newap;
					
				newap->prev = curap;
				nextblock = newap->data.next;
				curap = newap;
			}

			goto __pexit;
		}
	}

__pexit:	
	close(i_FD);
	free(blbuf);
}

/*
 * Get the defalt disklabel (=disk geometry) of the disk
 */
void
GetDefaultLabel(void)
{
	i_FD = opendisk(s_Device, O_RDONLY, buffer, MAXPATHLEN, 0);
	
	if (i_FD == -1)	{	
		printf("ERROR opening device. Make sure you are root.\n");

		exit(EXIT_FAILURE);
	}
	
	if (s_FullDev != NULL)
		free(s_FullDev);
		
	s_FullDev = strdup(buffer);
	
	/* 
	 * Get default disklabel
	 */
	if (ioctl(i_FD, DIOCGDEFLABEL, lp) < 0) {
		printf("ERROR reading default label. Make sure you are root.\n");
		exit(EXIT_FAILURE);
	}

	close(i_FD);
}

/*
 * Print disk data
 */
void
PrintDiskInfo(void)
{	
	printf("device: %s\n", s_FullDev);
	printf("bytes/sector: %d\n", lp->d_secsize);
	printf("sectors/track: %d\n", lp->d_nsectors);
	printf("tracks/cylinder: %d\n", lp->d_ntracks);
	printf("sectors/cylinder: %d\n", lp->d_secpercyl);
	printf("cylinders: %d\n", lp->d_ncylinders);
	printf("total sectors: %d\n", lp->d_secperunit);
	printf("rpm: %d\n", lp->d_rpm);
	printf("interleave: %d\n", lp->d_interleave);
	printf("trackskew: %d\n", lp->d_trackskew);
	printf("cylinderskew: %d\n", lp->d_cylskew);
	printf("headswitch: %d\t\t# microseconds\n", lp->d_headswitch);
	printf("track-to-track seek: %d\t# microseconds\n", lp->d_trkseek);
} 

/*
 * Command loop
 */
void 
CmdLoop(void)
{
	char *s;

	printf("\n*** Enter '?' for help\n\n");

	for (;;) {
		printf("partition> ");
		
		s = ReadStr();
	
		if (strlen(s) != 0) {
			const struct interactive *ip;
			
			for (ip = interactive; ip->cmd != NULL; ++ip) {
				if (s[0] == 'Q') {
					free(s);
					return;
				}
			
				if (strncmp(ip->cmd, s, strlen(ip->cmd)) == 0) {
					if (ip->ptr != NULL)
						ip->ptr();
					goto __free;
				}
				
				if (s[0] == 'c') {
					printf("Partition (c) is the whole disk and cannot be changed.\n");
					goto __free;
				}
				
				if ((s[0] >= 'a') && (s[0] <= 'p')) {
					MS_EditPartition(s[0]);
					goto __free;
				}
			}
		}
		
		printf("Unknown command\n");
		
	__free:	
		free(s);
	}
}

/*
 * Add a new partition
 */
amigapart *
AddPartition(char partname, amigapart *ap)
{
	amigapart *np;
	
	if (ap != NULL)	{
		if (ap->next != NULL)
			return NULL;
	}
		
	np = (amigapart *)malloc(sizeof(amigapart));
	memset(np, 0x00, sizeof(amigapart));
	
	memcpy(np->data.id, pa_id, 4);
	np->data.size_in_longs = 0x40;
	np->data.hostid = 7;
	np->data.next = ND;
	np->data.drivenamelen = 4;
	strcpy(np->data.drivename, "NEB0"); /* Fixup the '0' later */
	np->data.size_of_vector = 0x16;
	np->data.sizeblock = 0x80;
	np->data.surfaces = lp->d_ntracks;
	np->data.sectors_block = 1;
	np->data.blocks_track = lp->d_nsectors;
	np->data.numbuffer = 30;
	np->data.bufmemtype = 0;
	np->data.maxtransfer = 0x7fffffff;
	np->data.mask = 0xfffffffe;
	memcpy(np->data.dostype, pa_dos, 3); /* Will be overwritten */
	
	/* 
	 * Link partition. If first partition, link it
	 * directly from RDB
	 */
	if (ap != NULL)	{
		ap->next = np;
		np->prev = ap;
	} else {
		rdbp->partitionlist_lnk = np;
		np->prev = NULL;
	}
	
	return np;
}

/*
 * Edit a partition. Does NOT scan for overlapping partitions yet.
 */
int
EditPartition(amigapart *ap)
{
	uint32_t convhlp, dataoff, datasize, numbb, numbr, offset, size;
	int automount, boot;
	char *asel, *bsel, *fssel;
		
	convhlp = 0;
	dataoff = rdbp->data.locylinder;
	datasize = (rdbp->data.hicylinder - rdbp->data.locylinder + 1);
	numbb = 0;
	numbr = 0;
	offset = 0;
	size = 0;
	automount = 1;
	boot = 0;

	printf("*** Partition editor.\n");
	printf("*** Total data cylinders of drive is %d\n", datasize);
	printf("*** First data area cylinder is %d\n", dataoff);	
	printf("*** Note that all SIZE/OFFSET values are CYLINDERS, not BLOCKS!\n\n");

	printf("Start of partition (cylinder): ");
	offset = ReadUL();
	
	if (offset < dataoff) {
		printf("ERROR: Start is before data area start\n");
		return 0;
	}
	
	printf("Size of partition (cylinders): ");
	size = ReadUL();
	
	if (size > (datasize - offset))	{
		printf("ERROR: Partition is too large. Maximum allowed is %d \
			blocks\n", (datasize - offset));
		return 0;
	}
	
	if (size < 1)
		size = 1;
		
	/* See if this area is (partially) in use by another partition already */
	if (IsAreaUsed(offset, size, ap)) {
		printf("ERROR: Area %d-%d is in use by another partition already\n",
			offset, offset + size - 1);
		return 0;
	}		
	
	printf("Type of FS (r=NetBSD root, s=NetBSD swap, u=NetBSD other, <ENTER>=custom): ");
	fssel = ReadStr();
	
	switch (fssel[0]) {
		case 'r':
		case 'R':
			strcpy(ap->data.dostype, "NBR");
			ap->data.dostype[3] = 7;
			break;
			
		case 's':
		case 'S':
			strcpy(ap->data.dostype, "NBS");
			ap->data.dostype[3] = 1;
			break;
				
		case 'u':
		case 'U':
		case 'o':
		case 'O':
			strcpy(ap->data.dostype, "NBU");
			ap->data.dostype[3] = 7;
			break;				
			
		default:
			free(fssel);
			printf("Custom file system: Enter hex value (0x + 8 digits) e.g. 0x41424307: ");
			fssel = ReadStr();
		
			/*
			 * Someone should advance this input method !
			 * Hm ... perhaps a bit strage
			 */
			if (strlen(fssel) != 11) {
				printf("ERROR: Invalid value\n");
				free(fssel);
				
				return 0;
			}
	
			sscanf(fssel, "%x", &convhlp);
			
			ap->data.dostype[0] = (char)((convhlp & 0xff000000) >> 24);
			ap->data.dostype[1] = (char)((convhlp & 0x00ff0000) >> 16);
			ap->data.dostype[2] = (char)((convhlp & 0x0000ff00) >> 8);
			ap->data.dostype[3] = (char)(convhlp & 0x000000ff);		
			break;
	}
	
	printf("Mark partition bootable (y/n): ");
	bsel = ReadStr();
	
	if ((bsel[0] == 'y') || (bsel[0] == 'Y')) {
		boot = 1;
	
		printf("Number of boot blocks: ");
		numbb = ReadUL();
	}	
	
	printf("Number of reserved blocks: ");
	numbr = ReadUL();

	printf("Mark partition automount (y/n): ");
	asel = ReadStr();
	
	if ((asel[0] == 'y') || (asel[0] == 'Y'))
		automount = 0;

	/*
	 * All data collected. Fill structure. 
	 */
	ap->data.lowcyl = offset;
	ap->data.highcyl = offset + size-1;
	ap->data.bootblocks = numbb;
	ap->data.reserved = numbr;

	if (boot == 1)
		ap->data.flags |= 1;
	else
		ap->data.flags &= ~1;
		
	if (automount == 1)
		ap->data.flags |= 2;
	else
		ap->data.flags &= ~2;
	
	free(fssel);
	free(bsel);
	free(asel);
	
	return -1;	
}

/*
 * Delete a partition. Reorganizes chain.
 */
void
DeletePartition(amigapart *ap)
{
	if (ap == NULL)
		return;
		
	if (ap->prev != NULL)
		ap->prev->next = ap->next;
	else
		rdbp->partitionlist_lnk = ap->next;
		
	if (ap->next != NULL)
		ap->next->prev = ap->prev;
		
	free(ap);	
}

/*
 * Arrange blocks. Will calculate block pointers.
 * Fixes up drive names
 */
void
ArrangeBlocks(void)
{
	amigapart *ap;
	uint32_t blk;
	int cnt;
	
	ap = rdbp->partitionlist_lnk;
	blk = 1; /* 512 bytes */
	cnt = 0;

	if (ap != NULL) {
		rdbp->data.partitionlist = blk;
		
		while (ap != NULL) {
			/* Store PART blocks in a chain */
			ap->data.next = (ap->next != NULL) ? ++blk : ND;
			ap->data.drivename[3] = '0' + cnt++;
			ap = ap->next;
		}		
		
		rdbp->data.highrsdkblock = blk;
	} else {
		rdbp->data.highrsdkblock = 0;
	}
}

/*
 * Calculate checksums (RDSK+PART)
 */
int32_t
AmiChecksum(unsigned char *ptr)
{
	int32_t newsum;
	int i;
	
	newsum = 0;
	
	*(ptr + 8) = 0;
	*(ptr + 9) = 0;
	*(ptr + 10) = 0;
	*(ptr + 11) = 0;

	for (i = 0; i < 64; i++)
		newsum += LONG(ptr + i * 4);
		
	newsum = -newsum;			/* negation */
	
	return newsum;
}

/*
 * Calculate all checksums. Calling this is the last step before
 * writing to disk. Block pointers must have been fixed before
 * calling this, the data must be in BIG ENDIAN format already!
 */
void 
CalcChecksums(void)
{
	/* RDSK checksum */
	rdbp->data.checksum = 0;
	rdbp->data.checksum = htonl(AmiChecksum((unsigned char*)&rdbp->data));
	
	/* PART checksums */
	amigapart *ap = rdbp->partitionlist_lnk;
	
	while (ap != NULL)
	{
		ap->data.checksum = 0;
		ap->data.checksum = htonl(AmiChecksum((unsigned char*)&ap->data));
		
		ap = ap->next;
	}	
}

/*
 * Change the endianness of all data to the amiga format
 */
void
FormatToAmiga(void)
{
	amigapart *ap;
	
	rdbp->data.size_in_longs = htonl(rdbp->data.size_in_longs);
	rdbp->data.hostid = htonl(rdbp->data.hostid);
	rdbp->data.block_size = htonl(rdbp->data.block_size);
	rdbp->data.flags = htonl(rdbp->data.flags);
	rdbp->data.bad_blocklist = htonl(rdbp->data.bad_blocklist);
	rdbp->data.partitionlist = htonl(rdbp->data.partitionlist);
	rdbp->data.filesyshdrlist = htonl(rdbp->data.filesyshdrlist);
	rdbp->data.driveinit_code = htonl(rdbp->data.driveinit_code);
	rdbp->data.cylinders = htonl(rdbp->data.cylinders);
	rdbp->data.sectors = htonl(rdbp->data.sectors);
	rdbp->data.heads = htonl(rdbp->data.heads);
	rdbp->data.interleave = htonl(rdbp->data.interleave);
	rdbp->data.parking_zone = htonl(rdbp->data.parking_zone);
	rdbp->data.writeprecomp = htonl(rdbp->data.writeprecomp);
	rdbp->data.reducedwrite = htonl(rdbp->data.reducedwrite);
	rdbp->data.steprate = htonl(rdbp->data.steprate);
	rdbp->data.rdb_blocklo = htonl(rdbp->data.rdb_blocklo);
	rdbp->data.rdb_blockhi = htonl(rdbp->data.rdb_blockhi);
	rdbp->data.locylinder = htonl(rdbp->data.locylinder);
	rdbp->data.hicylinder = htonl(rdbp->data.hicylinder);
	rdbp->data.cylblocks = htonl(rdbp->data.cylblocks);
	rdbp->data.autoparkseconds = htonl(rdbp->data.autoparkseconds);
	rdbp->data.highrsdkblock = htonl(rdbp->data.highrsdkblock);
	
	ap = rdbp->partitionlist_lnk;

	while (ap != NULL)
	{
		ap->data.size_in_longs = htonl(ap->data.size_in_longs);
		ap->data.hostid = htonl(ap->data.hostid);
		ap->data.next = htonl(ap->data.next);
		ap->data.flags = htonl(ap->data.flags);
		ap->data.devflags = htonl(ap->data.devflags);
		ap->data.size_of_vector = htonl(ap->data.size_of_vector);
		ap->data.sizeblock = htonl(ap->data.sizeblock);
		ap->data.secorg = htonl(ap->data.secorg);
		ap->data.surfaces = htonl(ap->data.surfaces);
		ap->data.sectors_block = htonl(ap->data.sectors_block);
		ap->data.blocks_track = htonl(ap->data.blocks_track);
		ap->data.reserved = htonl(ap->data.reserved);
		ap->data.prealloc = htonl(ap->data.prealloc);
		ap->data.interleave = htonl(ap->data.interleave);
		ap->data.lowcyl = htonl(ap->data.lowcyl);
		ap->data.highcyl = htonl(ap->data.highcyl);
		ap->data.numbuffer = htonl(ap->data.numbuffer);
		ap->data.bufmemtype = htonl(ap->data.bufmemtype);
		ap->data.maxtransfer = htonl(ap->data.maxtransfer);
		ap->data.mask = htonl(ap->data.mask);
		ap->data.bootpri = htonl(ap->data.bootpri);
		ap->data.baud = htonl(ap->data.baud);
		ap->data.control = htonl(ap->data.control);
		ap->data.bootblocks = htonl(ap->data.bootblocks);
		
		ap = ap->next;
	}
}

/*
 * Convert the RDSK to local endianness
 */
void 
FormatRDBToLocal(void)
{
	rdbp->data.size_in_longs = ntohl(rdbp->data.size_in_longs);
	rdbp->data.hostid = ntohl(rdbp->data.hostid);
	rdbp->data.block_size = ntohl(rdbp->data.block_size);
	rdbp->data.flags = ntohl(rdbp->data.flags);
	rdbp->data.bad_blocklist = ntohl(rdbp->data.bad_blocklist);
	rdbp->data.partitionlist = ntohl(rdbp->data.partitionlist);
	rdbp->data.filesyshdrlist = ntohl(rdbp->data.filesyshdrlist);
	rdbp->data.driveinit_code = ntohl(rdbp->data.driveinit_code);
	rdbp->data.cylinders = ntohl(rdbp->data.cylinders);
	rdbp->data.sectors = ntohl(rdbp->data.sectors);
	rdbp->data.heads = ntohl(rdbp->data.heads);
	rdbp->data.interleave = ntohl(rdbp->data.interleave);
	rdbp->data.parking_zone = ntohl(rdbp->data.parking_zone);
	rdbp->data.writeprecomp = ntohl(rdbp->data.writeprecomp);
	rdbp->data.reducedwrite = ntohl(rdbp->data.reducedwrite);
	rdbp->data.steprate = ntohl(rdbp->data.steprate);
	rdbp->data.rdb_blocklo = ntohl(rdbp->data.rdb_blocklo);
	rdbp->data.rdb_blockhi = ntohl(rdbp->data.rdb_blockhi);
	rdbp->data.locylinder = ntohl(rdbp->data.locylinder);
	rdbp->data.hicylinder = ntohl(rdbp->data.hicylinder);
	rdbp->data.cylblocks = ntohl(rdbp->data.cylblocks);
	rdbp->data.autoparkseconds = ntohl(rdbp->data.autoparkseconds);
	rdbp->data.highrsdkblock = ntohl(rdbp->data.highrsdkblock);
}

/*
 * Format the given partition to local endianness
 */
void
FormatPARTToLocal(amigapart *ap)
{
	if (ap != NULL)	{
		ap->data.size_in_longs = ntohl(ap->data.size_in_longs);
		ap->data.hostid = ntohl(ap->data.hostid);
		ap->data.next = ntohl(ap->data.next);
		ap->data.flags = ntohl(ap->data.flags);
		ap->data.devflags = ntohl(ap->data.devflags);
		ap->data.size_of_vector = ntohl(ap->data.size_of_vector);
		ap->data.sizeblock = ntohl(ap->data.sizeblock);
		ap->data.secorg = ntohl(ap->data.secorg);
		ap->data.surfaces = ntohl(ap->data.surfaces);
		ap->data.sectors_block = ntohl(ap->data.sectors_block);
		ap->data.blocks_track = ntohl(ap->data.blocks_track);
		ap->data.reserved = ntohl(ap->data.reserved);
		ap->data.prealloc = ntohl(ap->data.prealloc);
		ap->data.interleave = ntohl(ap->data.interleave);
		ap->data.lowcyl = ntohl(ap->data.lowcyl);
		ap->data.highcyl = ntohl(ap->data.highcyl);
		ap->data.numbuffer = ntohl(ap->data.numbuffer);
		ap->data.bufmemtype = ntohl(ap->data.bufmemtype);
		ap->data.maxtransfer = ntohl(ap->data.maxtransfer);
		ap->data.mask = ntohl(ap->data.mask);
		ap->data.bootpri = ntohl(ap->data.bootpri);
		ap->data.baud = ntohl(ap->data.baud);
		ap->data.control = ntohl(ap->data.control);
		ap->data.bootblocks = ntohl(ap->data.bootblocks);
	}
}

/*
 * Change the endianness of all data to the local format
 */
void
FormatToLocal(void)
{
	amigapart *ap;
	
	FormatRDBToLocal();
	
	ap = rdbp->partitionlist_lnk;
	
	while (ap != NULL) {
		FormatPARTToLocal(ap);
		ap = ap->next;
	}
}

/*
 * Write tables to disk.
 * Builds a single buffer with all blocks.
 * Assumes PART blocks to be stored continuous.
 */
void 
Write(uint32_t bsize, uint32_t firstpartoffset, uint32_t blockstep)
{
	amigapart *ap;
	unsigned char *wb, *wb_p;
	
	wb_p = (unsigned char *)malloc(bsize);
	wb = wb_p;

	memset(wb, 0x00, bsize);	
	memcpy(wb, &rdbp->data, 256);

	ap = rdbp->partitionlist_lnk;
	wb = wb_p + firstpartoffset;
	
	while (ap != NULL) {
		memcpy(wb, &ap->data, 256);		
		wb += blockstep;
	
		ap = ap->next;
	}
	
	/* Write label to disk */
	i_FD = opendisk(s_Device, O_WRONLY, buffer, MAXPATHLEN, 0);
	
	if (i_FD == -1)	{	
		printf("ERROR opening device. Make sure you are root.\n");

		exit(EXIT_FAILURE);
	}
	
	write(i_FD, wb_p, bsize);
	close(i_FD);
	
	/* 
	 * DEBUG: Dump to file. Should write this to disk in final version!
	 */
/*	FILE *fp = fopen("disk.bin", "wb+");
	fwrite(wb_p, bsize, 1, fp);
	fclose(fp);
	
	free(wb_p);*/
}

/*
 * Check if a given area is in use by any partition already.
 * Do not consider self-overlapping a "violation".
 */
int
IsAreaUsed(uint32_t offset, uint32_t size, struct amigapart *self)
{
	amigapart *ap;
	uint32_t a_start, a_end;	
	
	ap = rdbp->partitionlist_lnk;
	a_start = offset;
	a_end = offset + size - 1;
	
	while (ap != NULL) {
		if (ap != self) {
			if (BETWEEN(a_start, ap->data.lowcyl, ap->data.highcyl)
				|| BETWEEN(a_end, ap->data.lowcyl, ap->data.highcyl)
				|| SPAN(a_start, a_end, ap->data.lowcyl, ap->data.highcyl))	{
			
				return 1;
			}
		}
		
		ap = ap->next;
	}
	
	return 0;	
}


/*
 * Menu command: Print help text
 */
void
M_PrintHelp(void)
{
	const struct interactive *ip;
	
	for (ip = interactive; ip->cmd != NULL; ++ip)
		printf("\t%s\t%s\n", ip->cmd, ip->hlptext);
}

/*
 * Menu command: Reset RDB and all partitions
 */
void 
M_ClearRDB(void)
{
	ResetRDB();
}

/*
 * Menu command: Print all partitions
 */
void
M_PrintPart(void)
{
	amigapart *ap;
	uint32_t cvb_u;
	int cnt;
	
	ap = rdbp->partitionlist_lnk;
	cvb_u = 0;
	cnt = 0;

	if (ap == NULL)
		printf("No partitions defined yet. Showing only whole disk (c)\n\n");
		
	printf("#          size      offset     system mount? boot?\n");

	while (ap != NULL) {
		if (cnt == 2)
			P_WholeDisk();
		else {
			/* Convert DOS-Type to a single uint */
			cvb_u = 0;
			
			cvb_u |= (unsigned char)ap->data.dostype[0];
			cvb_u <<= 8;
			cvb_u |= (unsigned char)ap->data.dostype[1];
			cvb_u <<= 8;
			cvb_u |= (unsigned char)ap->data.dostype[2];
			cvb_u <<= 8;
			cvb_u |= (unsigned char)ap->data.dostype[3];
			
			P_FmtLine(cnt, (ap->data.highcyl-ap->data.lowcyl+1), \
				ap->data.lowcyl, cvb_u, \
				ap->data.flags & 2 ? "NO" : "YES", \
				ap->data.flags & 1 ? "YES" : "NO");

			ap = ap->next;
		}
		
		cnt++;
	}
	
	if (cnt <= 2)
		P_WholeDisk();
}

/*
 * Menu command: Delete a partition
 */
void
M_DeletePart(void)
{
	amigapart *ap;
	int pnum, cnt;
	char *wpd;
	
	printf("Delete partition: ");
	wpd = ReadStr();
	
	if (strlen(wpd) != 1) {
		if (wpd[0] == 'c') {
			printf("ERROR: Partition \"c\" is the whole disk. Cannot delete.\n");

			goto __freeup;
		}
	
		pnum = wpd[0] - 'a';
		
		if (pnum < 0)
			printf("ERROR: Invalid parition\n");
		else {
			/* Correct for partition (c) */
			if (pnum > 2)
				pnum--;
					
			ap = rdbp->partitionlist_lnk;
			cnt = 0;
			
			while (ap != NULL) {
				if (cnt == pnum)
					break;
					
				ap = ap->next;
				cnt++;
			}
			
			if (ap != NULL)
				DeletePartition(ap);
			else
				printf("ERROR: Invalid partition specified");
		}
	}
	
__freeup:
	free(wpd);	
}

/*
 * Menu command: Write data to disk
 */
void
M_Write(void)
{
	int blockstep, bsize, firstpartoffset;
	
	ArrangeBlocks();
	
	bsize = rdbp->data.locylinder * rdbp->data.cylblocks * \
		rdbp->data.block_size;
	firstpartoffset = (rdbp->data.partitionlist * rdbp->data.block_size);
	blockstep = rdbp->data.block_size;	
	
	FormatToAmiga();	
	CalcChecksums();
	Write(bsize, firstpartoffset, blockstep);
	FormatToLocal();
}

/*
 * Menu command: Edit a given partition
 */
void
MS_EditPartition(char c)
{
	/* c ist guaranteed not to be partition "c" (whole disk) */
	amigapart *ap, *ap_p;
	int dnum, stepcnt;
	
	ap = rdbp->partitionlist_lnk;
	ap_p = rdbp->partitionlist_lnk;
	dnum = c - 'a';
	stepcnt = 0;

	/* Adjust for partitions beyond "c" */
	if (dnum >= 2)
		dnum--;
	
	while (ap != NULL) {
		if (stepcnt == dnum) {
			/* Partition found, edit */
			EditPartition(ap);
			
			return;
		}
		
		stepcnt++;
		ap_p = ap;
		ap = ap->next;
	}	
	
	if (stepcnt == dnum) {
		/* Desired partition is last+1, so add it */
		ap = AddPartition('A' + dnum, ap_p);
		
		if (EditPartition(ap) == 0)
			DeletePartition(ap);
	}
}

/*
 * Format a partition line for printing
 */
void
P_FmtLine(int cnt, uint32_t size, uint32_t offset, uint32_t system,
	const char *mount, const char *boot)
{
	int i;
	char buf[1024], s1[3], s2[12], s3[12], s4[11], s5[7], s6[6];
	
	memset(buf, 0x00, 1024);
	memset(s1, 0x00, 3);
	memset(s2, 0x00, 12);
	memset(s3, 0x00, 12);
	memset(s4, 0x00, 11);
	memset(s5, 0x00, 7);
	memset(s6, 0x00, 6);
	
	s1[0] = 'a' + cnt;
	s1[1] = ':';
	
	sprintf(buf, "%d", size);	
	for (i = 0; i < 11 - strlen(buf); i++)
		s2[i] = ' ';		
	strcat(s2, buf);

	sprintf(buf, "%d", offset);	
	for (i = 0; i < 11 - strlen(buf); i++)
		s3[i] = ' ';		
	strcat(s3, buf);
	
	sprintf(buf, "%x", system);
	for (i = 0; i < 8 - strlen(buf); i++)
		s4[i] = ' ';
	strcat(s4, "0x");
	strcat(s4, buf);
	
	for (i = 0; i < 6 - strlen(mount); i++)
		s5[i] = ' ';
	strcat(s5, mount);

	for (i = 0; i < 5 - strlen(boot); i++)
		s6[i] = ' ';
	strcat(s6, boot);

	printf(" %s %s %s %s %s %s\n", s1, s2, s3, s4, s5, s6);
}

/*
 * Print format line for the special "c" partition
 */
void
P_WholeDisk(void)
{
	P_FmtLine(2, rdbp->data.cylinders, 0, 0, "-", "-");
}

/*
 * neb-hdtoolbox: Program to create initial partition table for
 * the amiga-port.
 */
int 
main(int argc, char *argv[])
{
	setprogname(argv[0]);
	
	if (argc < 2)
		usage();
	
	/* Quick hack to filter -h and -? ... need to rework this */
	if (((strcmp(argv[1], "-h") == 0) || ((strcmp(argv[1], "-?")) == 0)))
		usage();
	
	/* Data setup */
	s_Device = argv[1];
	s_FullDev = NULL;
	lp = (struct disklabel *)malloc(sizeof(struct disklabel));
	memset(lp, 0x00, sizeof(struct disklabel));
	rdbp = (rdb *)malloc(sizeof(rdb));
	memset(rdbp, 0x00, sizeof(rdb));
	
	/* 
	 * First acquire the default label for the disk, then run
	 * the command loop
	 */
	GetDefaultLabel();
	ResetRDB();
	LoadLabel();
	PrintDiskInfo();
	CmdLoop();
	
	printf("\n");
	
	/* Cleanup */
	ResetRDB();
	free(rdbp);
	free(s_FullDev);

	exit(EXIT_SUCCESS);
}
